--[[ 
  @file       ofollow.lua
  @brief      main lua file for oFollow

  @author     rmcinnis
  @date       march 5, 2013
  @copyright  Solid ICE Technologies.  All rights reserved.
              this file may be distributed so long as it remains unaltered
              if this file is posted to a web site, credit must be given to me along with a link to my web page
              no code in this file may be used in other works without expressed permission  
]]--
local _addonName, OFaddon = ... ;
if (OFaddon.table == nil) then
  OFaddon.table = {} ;
end
local tbl = OFaddon.table ;

local OF_MAJOR             = 0 ;
local OF_MINOR             = 2 ;
local OF_REVISION          = 7 ;
local OF_VERSION           = tostring(OF_MAJOR) ..".".. tostring(OF_MINOR) ..".".. OF_REVISION ;
local _debug               = nil ;
local _adjust              = nil ;

if (OF == nil) then
  OF = { pixel_sz = 1, is_master = 0 } ;
end
local of = {} ;
of.dots = {} ;

SLASH_OFOLLOW1 = '/of' ;
SlashCmdList["OFOLLOW"] = function (msg, editbox)
  if (msg == nil) or (msg == "") then
    of.config() ;
    return
  end
  local arg1 = msg ;
  local opts = nil ;
  if (msg ~= nil) and (msg:find(" ") ~= nil) then
    arg1 = msg:sub(1,msg:find(" ")-1) ;
    opts = msg:sub(msg:find(" ")+1,-1) ;
  end
  if (of.options[ arg1 ] ~= nil) then
    of.options[ arg1 ]( opts ) ;
  end
end

-------------------------------------------------------------------------------
function tbl.new()
  if (tbl.__pool == nil) then
    tbl.__pool = {} ;
  end
  local t = next(tbl.__pool) ;
  if t then
    tbl.__pool[t] = nil ;
  else
    t = {} 
    tbl._count = (tbl._count or 0) + 1 ;
  end
  return t ;
end

function tbl.delete(t)
  if (t) then
    tbl.clear(t) ;
    tbl.__pool[t] = true ;
    tbl._count = (tbl._count or 0) - 1 ;
  end
  return nil ;
end

function tbl.clear(t, deep)
  if (t) then
    if (deep) then
      for k in pairs(t) do 
        if (type(t[k]) == "table") then
          tbl.delete( tbl.clear( t[k], deep ) ) ;    -- clear sub tables and be able to reclaim
        end
        t[k] = nil ;
      end
    else
      wipe(t) ;
    end
  end
  return t ;
end

function tbl.size(t)
  if (t == nil) then
    return nil ;
  end
  local n = 0 ;
  for i,v in pairs(t) do
    n = n + 1 ;
  end
  return n ;
end

function tbl.fill( t, ... )
  if (t ~= nil) then
    tbl.clear(t) ;
    for i = 1,select('#', ... ) do
      t[i] = select(i, ...) ;
    end
  end
end

function tbl.fill_match( t, str, ch )
  local n = 0 ;
  wipe(t) ;
  local p1 = 0 ;
  local p2 = 1 ;
  if (str == nil) then return ; end   
  for i=1,#str do
    if (str:sub(i,i) == ch) then
      p2 = i ;
      n = n + 1 ;
      t[n] = str:sub(p1+1,p2-1) ;
      p1 = p2 ;
    end
  end
  if (p1 == p2) and (p2 ~= #str) then
    n = n + 1 ;
    t[n] = str:sub(p1+1,-1) ;
  end
  return n ;
end
-------------------------------------------------------------------------------

function OFollow_OnLoad(self)
  of.ui = self ;
  of.ui:Hide() ;
  of.register_events() ;
end

function of.CreateFrame( type, name, parent, template )
  local f = CreateFrame( type, name, parent, template ) ;
  if (parent ~= nil) then
    f:SetFrameLevel( parent:GetFrameLevel() + 1 ) ;
  end
  return f ;
end

function of.closebox( parent, on_close )
  of.nthings = (of.nthings or 0) + 1 ;
  closepb = of.CreateFrame("Button","$parent".. "Close".. of.nthings, parent, "UIPanelCloseButton") ;
  closepb:SetWidth(25) ;
  closepb:SetHeight(25) ;
  closepb:SetPoint("TOPRIGHT", parent, "TOPRIGHT", -3, -3) ;
  if (on_close == nil) then
    closepb:SetScript("OnClick", function(self) self:GetParent():Hide() ; end) ;
  else
    closepb:SetScript("OnClick", on_close ) ;
  end
  closepb:Show() ;
  return closepb ;
end

function of.button( parent, x, y, cx, cy, text, on_click_func )
  local button = of.CreateFrame("Button", "OF.Button", parent, "UIPanelButtonTemplate")

  button:SetWidth(cx)
  button:SetHeight(cy)
  button:SetNormalFontObject("GameFontNormalSmall")
  button:SetHighlightFontObject("GameFontHighlightSmall")
  button:SetDisabledFontObject("GameFontDisableSmall")
  button:SetText( text or "" )
  button:SetScript("OnClick", on_click_func )
  of.moveto( button, x, y ) ;
  button:Show() 

  return button ;
end

function of.editline( parent, x, y, cx, cy, max_chars )
  of.nthings = (of.nthings or 0) + 1 ;
  local e = of.CreateFrame("EditBox", "OF.Editline".. tostring(of.nthings), parent, "InputBoxTemplate" )
  e:SetPoint("TOPLEFT", parent, "TOPLEFT", 0,0 ) ; 
  e:SetText( "" ) ;
  e:SetAutoFocus(false)
  e:SetFontObject("GameFontNormal")
  e:SetMaxLetters(max_chars or 30)
  e:SetCursorPosition(0) ;
  e:SetTextColor( 0.9, 0.9, 0.9, 1 ) ;
  e.str = "" ;
  e:SetScript( "OnTextChanged"  , function(self) self.str = self:GetText() or "" ; if (self.func ~= nil) then self.func(self.str) ; end end ) ;
  e:SetScript( "OnEscapePressed", function(self) self:ClearFocus() end ) ;
  of.setpos( e, x, y, cx, cy ) ;
  e:Show() ;
  return e ;
end

function of.panel( parent, name, x, y, cx, cy, no_texture )
  local f = CreateFrame("FRAME", "$parent".. name, parent )
  f:SetWidth (cx) ; -- Set these to whatever height/width is needed 
  f:SetHeight(cy) ; -- for your Texture
  f:SetBackdropColor(0.2,0.2,0.2,1.0);

  if (not no_texture) then
    local t = f:CreateTexture(nil,"BACKGROUND") ;
    t:SetAllPoints(f) ;
    t:SetDrawLayer("BACKGROUND") ;
    f.texture = t ;
  end

  f:SetPoint( "TOPLEFT", x, -1 * y) ;
  return f ;
end

function of.setpos( f, x, y, cx, cy )
  if (f) then
    of.moveto( f, x, y ) ;
    if (cx ~= nil) and (cx > 0) then
      f:SetWidth(cx) ;
    end
    if (cy ~= nil) and (cy > 0) then
      f:SetHeight(cy) ;
    end
  end
  return f ;
end

function of.label( parent, x, y, cx, cy, text, justify_v, justify_h, font )
  local label = parent:CreateFontString( nil, "BACKGROUND", font or "GameFontNormalSmall")
  label:SetWidth( cx ) ;
  label:SetHeight(cy or 25)
  label:SetJustifyV( justify_v or "CENTER" )
  label:SetJustifyH( justify_h or "LEFT" )
  label:SetText( text )
  label:Show() 
  of.moveto( label, x, y ) ;
  return label ;
end

function of.make_config()
  if (of._config) then
    return of._config ;
  end
  
  local parent = of.panel( UIParent, "oFollowConfig", 100, 100, 250, 225 ) ;
  parent:ClearAllPoints() ;
  parent:SetPoint("CENTER", UIParent, "CENTER", 0, 300) ;
  parent:SetMovable(true) ;
  parent:EnableMouse(true)
  parent:RegisterForDrag("LeftButton")
  parent:SetScript("OnDragStart", parent.StartMoving)
  parent:SetScript("OnDragStop", parent.StopMovingOrSizing)
  parent:SetBackdrop({ bgFile="Interface/Tooltips/UI-Tooltip-Background", 
                       edgeFile="Interface/Tooltips/UI-Tooltip-Border", 
                       tile=true, tileSize = 16, edgeSize = 16,
                       insets = { left = 1, right = 1, top = 1, bottom = 1 }
                     })
  parent:SetBackdropColor( 0.2, 0.2, 0.2, 1.0 ) ;
  of.closebox( parent ) ;

  -- title
  local x, y, cx, cy ;
  x = 10 ;
  y = 10 ;
  cx = 150 ;
  cy = 30 ;
  parent.title = of.label( parent, x, y, cx, cy, "oFollow v".. OF_VERSION ) ;
  parent.title:SetFont( "Fonts\\FRIZQT__.TTF", 12, "") ;

  -- instructions  
  y = y + cy + 10 ;
  parent.title = of.label( parent, x, y, parent:GetWidth() - 2*x, cy, "Please enter the dimensions of your wows" ) ;

  -- master
  x = x + 10 ;
  y = y + cy ;
  of.label( parent, x, y, cx, cy, "master" ) ;
  
  parent._master_cx = of.editline( parent, x +  70, y, 50, cy, 4 ) ;
  of.label( parent, x + 125, y, cx, cy, "x" ) ;
  parent._master_cy = of.editline( parent, x + 145, y, 50, cy, 4 ) ;
  
  -- slave
  y = y + cy ;
  of.label( parent, x, y, cx, cy, "slave" ) ;
  
  parent._slave_cx = of.editline( parent, x +  70, y, 50, cy, 4 ) ;
  of.label( parent, x + 125, y, cx, cy, "x" ) ;
  parent._slave_cy = of.editline( parent, x + 145, y, 50, cy, 4 ) ;
  
  -- buttons  
  y = parent:GetHeight() - 40 ;
  x = parent:GetWidth () - 75 ;
  parent.ok_but = of.button( parent, x, y, 60, 30, "save", function(self) self:GetParent():Hide() ; of.save_config( self:GetParent() ) ; end ) ; 
    
  of._config = parent ;
  return parent ;
end

function of.save_config( f )
  OF.master_cx = tonumber( f._master_cx:GetText() or "1920" ) or 1920 ;
  OF.master_cy = tonumber( f._master_cy:GetText() or "1080" ) or 1080 ;
  OF.slave_cx  = tonumber( f._slave_cx :GetText() or "1920" ) or 1920 ;
  OF.slave_cy  = tonumber( f._slave_cy :GetText() or "1080" ) or 1080 ;
  
  of.load_config() ;
end

function of.load_config()
  if (OF.is_master == 1) then 
    of.set_dimension( OF.master_cx, OF.master_cy ) ;
  else
    of.set_dimension( OF.slave_cx , OF.slave_cy  ) ;
  end
end

function of.set_dimension( cx, cy )
  cx = tonumber(cx) or 0 ;
  cy = tonumber(cy) or 0 ;
  if (cx == 0) then cx = 1 ; end
  if (cy == 0) then cy = 1 ; end
  if (cx < 100) or (cy < 100) then
    of.reset_config() ;
  end
  of.__cx = cx ;
  of.__cy = cy ;
  of.on_pixel_size( (GetScreenWidth ()/cx), (GetScreenHeight()/cy) ) ;
end

function of.reset_config()
  OF.master_cx = 1920 ;
  OF.master_cy = 1080 ;
  OF.slave_cx  = 1920 ;
  OF.slave_cy  = 1080 ;
  
  of.load_config() ;
end

function of.config()
  local f = of.make_config() ;
  
  -- init fields
  f._master_cx:SetText( tostring(OF.master_cx or "") ) ;
  f._master_cy:SetText( tostring(OF.master_cy or "") ) ;
  f._slave_cx :SetText( tostring(OF.slave_cx  or "") ) ; 
  f._slave_cy :SetText( tostring(OF.slave_cy  or "") ) ;
  
  tinsert( UISpecialFrames, f:GetName() ) ;
  f:Show() ;
end

function of.create_timers()
  of.timers = {} ;
--  of.timer_clear() ;
end

function of.debug_report( ... )
  if (_debug) then
    print( ... ) ;
  end
end

function of.dump_stats()
  for i=0,10,1 do
    local f = of.ui.dotframe.dots[i] ;
    print( i ..".  ".. floor(f:GetLeft()) ..", ".. floor(f:GetTop()) .."  ".. f:GetWidth() .." x ".. f:GetHeight() .." " ) ;
  end
  print( "pixel_cx: ".. tostring(OF.pixel_cx) ) ;
  print( "pixel_cy: ".. tostring(OF.pixel_cy) ) ;
  
  print( "master: ".. tostring(OF.master_cx) .." x ".. tostring(OF.master_cy) ) ;
  print( "slave : ".. tostring(OF.slave_cx)  .." x ".. tostring(OF.slave_cy ) ) ;
  print( "master: ".. tostring(OF.is_master) ) ;
end

function of.hook_options()
  of.options = {} ;
  of.options[ "break"       ] = of.break_follow ;
  of.options[ "combat"      ] = of.combat_starts ;
  of.options[ "config"      ] = of.config ;
  of.options[ "debug"       ] = of.toggle_debug ;
  of.options[ "fmt"         ] = of.fmt ;
  of.options[ "follow"      ] = of.follow ;
  of.options[ "master"      ] = of.master ;
  of.options[ "off"         ] = of.off ;
  of.options[ "on"          ] = of.on ;
  of.options[ "reset"       ] = of.reset_config ;
  of.options[ "slave"       ] = of.slave ;
  of.options[ "stats"       ] = of.dump_stats ;
  of.options[ "test"        ] = of.toggle_test ;
  of.options[ "timers"      ] = of.timer_dump ;
  of.options[ "-v"          ] = of.version ;
  of.options[ "?"           ] = of.usage ;
end

function of.toggle_test( opt )
  if (opt == nil) then
    if (_adjust) then
      _adjust = nil ;
    else
      _adjust = 1 ;
    end
  else
    _adjust = tonumber(opt) ;
  end
  print( "test mode: ".. tostring(_adjust) ) ;
end

function of.init()
  OF.pixel_cx  = OF.pixel_cx or 1 ;
  OF.pixel_cy  = OF.pixel_cy or 1 ;
  OF.is_master = OF.is_master or 0 ;
  if (OF.aspect == nil) then
    OF.aspect = { ["slave" ] = { cx = 1.0, cy = 1.0 },
                  ["master"] = { cx = 1.0, cy = 1.0 },
                } ;
  end
  of.active    = true ;
  
  of.hook_options() ;
  of.create_timers() ;
  
  of.dots.create() ;
  of.version() ;  
  
  -- timed events
  of.timer( "xmit_status", 0.1, of.xmit_status, true ) ;  
  of.timer( "of_load"    , 1.0, of.load_config, nil ) ;  
end

function of.fmt()
  of.is_fmt = 1 ;
end

function of.moveto( f, x, y ) 
  if (y >= 0) then
    if (x >= 0) then 
      f:SetPoint("TOPLEFT",f:GetParent(),"TOPLEFT", x, -1 * y)
    else
      f:SetPoint("TOPRIGHT",f:GetParent(),"TOPRIGHT", x, -1 * y)
    end
  else
    if (x >= 0) then 
      f:SetPoint("BOTTOMLEFT",f:GetParent(),"BOTTOMLEFT", x, -1 * y)
    else
      f:SetPoint("BOTTOMRIGHT",f:GetParent(),"BOTTOMRIGHT", x, -1 * y)
    end
  end
end

function of.off()
  of.active = nil ;
  print( "OFollow inactive" ) ;
end

function of.on()
  of.active = true ;
  print( "OFollow active" ) ;
end

function of.on_addon_loaded( name )
  if (name == _addonName) then
    of.init() ;
  end
end

function of.on_event( self,event,... )
  if (of.msg_handler[event] ~= nil) then
    of.msg_handler[event]( ... ) ;
  end  
end

function of.on_pixel_size( opt_cx, opt_cy )
  OF.pixel_cx = tonumber(opt_cx) or 1 ;
  OF.pixel_cy = tonumber(opt_cy) or 1 ;
  if (OF.pixel_cx > 100) or (OF.pixel_cy > 100) then
    of.reset_config() ;
  else
    of.update_scale( OF.pixel_cx, OF.pixel_cy ) ;
  end
end

--
--  /of set dim master 1453 880
--
local _vars = {} ;
function of.on_timer( now )
  of.timer_trigger( now ) ;
end

function of.register_events()
  of.msg_handler = {} ;
  of.msg_handler[ "ADDON_LOADED" ] = of.on_addon_loaded ;

  of.ui:RegisterEvent( "ADDON_LOADED" ) ;
  of.ui:SetScript( "OnEvent", of.on_event ) ;
  if (RegisterAddonMessagePrefix( "OF" ) ~= true) then
    print( "Error:  unable to register addon prefix: OF" ) ;
  end
end

function of.break_follow()
  of.is_follow = 0 ;
end

function of.combat_starts()
  of.is_combat = 1 ; 
  of.is_follow = 0 ;
end

function of.follow()
  of.is_follow = 1 ;
  of.is_combat = 0 ; 
end

function of.master()
  OF.is_master = 1 ;
  of.load_config() ;
end

function of.slave()
  OF.is_master = 0 ;
  of.load_config() ;
end

function of.xmit_self()
  of.dots.xmit( 99, UnitName("player") .."@".. GetRealmName() ..".".. tostring(select(1,UnitFactionGroup("player"))) ) ;
end

local ticker = 0 ;
function of.xmit_status()
  ticker = ticker + 1 ;
  if ((ticker % 10) == 0) then
    of.xmit_self() ;
    return ;
  end
-- suspecting ofollow of triggering lots of WORLD_MAP_UPDATE events
--  SetMapToCurrentZone() 

  local zone = GetCurrentMapAreaID() ;
  if (of.zone ~= zone) then
    SetMapToCurrentZone() 
  end
  of.zone       = zone ;
  of.x, of.y    = GetPlayerMapPosition( "player" ) ;
  of.facing     = math.floor(GetPlayerFacing() * 180.0 / math.pi) ;
  of.health     = math.floor(100.0 * (UnitHealth("player") / UnitHealthMax("player"))) ;
  of.is_mounted = IsMounted() and 1 or 0 ;
  of.is_dead    = UnitIsDeadOrGhost("player") or 0 ;
  if (of.is_dead == 1) then
    of.is_follow = 0 ; 
  end
  if (UnitHealthMax("target") > 0) then
    of.target_health = math.floor(100.0 * UnitHealth("target") / UnitHealthMax("target"))
  else
    of.target_health = 0 
  end
  of.dots.xmit( 10, tostring(of.x) ..",".. tostring(of.y) ..",".. tostring(of.zone) ..","..
                    tostring(of.facing) ..",".. 
                    tostring(of.health) ..",".. tostring(of.target_health) ..",".. 
                    tostring(of.is_follow or 0) .. tostring(OF.is_master or 0) ..
                    tostring(of.is_mounted or 0) .. tostring(of.is_dead or 0) ..
                    tostring(of.is_combat or 0) .. tostring(of.is_debug or 0) ..
                    tostring(of.is_fmt or 0)
              ) ;
end

function of.debug_off()
  _debug = nil ;
  of.is_debug = 0 ;
  print( "ofollow:  debug off" ) ;
end

function of.debug_on()
  _debug = true ;
  of.is_debug = 1 ;
  print( "ofollow:  debug on" ) ;
end

function of.toggle_debug( opt )
  if (opt == nil) then
    if (_debug) then
      of.debug_off() ;
    else
      of.debug_on() ;
    end
  elseif (opt == "off") then
      of.debug_off() ;
  elseif (opt == "on") then
      of.debug_on() ;
  end
end

function of.usage() 
  of.version() ;
  print( "possible commands:" ) ;
  print( "  break           break follow" ) ;
  print( "  combat          combat mode" ) ;
  print( "  config          adjust oFollow settings" ) ;
  print( "  debug           allow ofollow to work outside of bgs" ) ;
  print( "  follow          start follow" ) ;
  print( "  master          set the current wow as the master" ) ;
  print( "  slave           set the current wow to slave mode" ) ;
end

function of.version() 
  print( "oFollow v".. OF_VERSION ) ;
end

--------------------------------------------------------------------------
-- dots
--------------------------------------------------------------------------
--
-- pixel perfect utilities
-- http://nclabs.org/articles/2
--
local function scale(numpixels)
    local uiscale = .64 ;
    local resoheight = (of.__cy or 768) ;
    local px = 768/resoheight/uiscale ;
    
    return floor( px * floor(numpixels + .5)) ;
end

local function size(frame, width, height)
  if not height then
    height = width
  end
  frame:SetSize(scale(width), scale(height))
end

local function point(obj, arg1, arg2, arg3, arg4, arg5)
  -- anyone has a more elegant way for this?
  if type(arg1)=="number" then arg1 = scale(arg1) end
  if type(arg2)=="number" then arg2 = scale(arg2) end
  if type(arg3)=="number" then arg3 = scale(arg3) end
  if type(arg4)=="number" then arg4 = scale(arg4) end
  if type(arg5)=="number" then arg5 = scale(arg5) end
  obj:SetPoint(arg1, arg2, arg3, arg4, arg5)
end

local function inject(obj)
    local mt = getmetatable(obj).__index
    mt.Size = size
    mt.Point = point
end

function of.update_scale( pixel_cx, pixel_cy )
  local prev = nil ;
  for i=0,106,1 do
    local f = of.ui.dotframe.dots[i] ;
    local eScale = f:GetParent():GetEffectiveScale() ;
    f:SetSize( pixel_cx, pixel_cy ) ;
    if (prev == nil) then
      of.moveto( f, i, 0 ) ;
    else
      f:SetPoint( "TOPLEFT", prev:GetName(), "TOPRIGHT", 0, 0 ) ;
    end
    prev = f ;
  end
end

--
-- end of pixel perfect utilities
--

local ndots = 0 ;
function of.dots.create_dot( parent, x, y, cx, cy, prev_frame ) 
  ndots = ndots + 1 ;
  local n = "Dot".. ndots ;
  local f = CreateFrame( "Frame", n, parent ) ;
  local val = 5 ;
  hooksecurefunc( f, "SetScale", function( self, scale )
                    if (self.fixingScale) then return end
                    self.fixingScale = true ;
                    -- check the size/scale and change them if desired
                    self.fixingScale = false
                  end
                ) ;

  f:SetAlpha( 1.0 ) ;
  f.val = 0 ;
  n = "DotTxt".. ndots ;
  f.texture = f:CreateTexture(n) ;
  f.texture:SetAllPoints(f) ;
  if (f.texture.SetColorTexture == nil) then
    f.texture:SetTexture( 0.2, 0.2, 0.0, 1 ) ; -- allowing for private servers that have not updated
  else
    f.texture:SetColorTexture( 0.2, 0.2, 0.0, 1 ) ;
  end
  f:SetFrameStrata( "TOOLTIP" ) ;
  f:SetFrameLevel( 128 ) ; -- 129 is the max; 100 should be good enough to make it topmost

  f:SetSize( cx, cy ) ;
  if (prev_frame == nil) then
    of.moveto( f, scale(x), scale(y) ) ;
  else
    f:SetPoint( "TOPLEFT", prev_frame:GetName(), "TOPRIGHT", 0, 0 ) ;
  end

  f:Show() ;
  return f ;
end

function of.dots.create() 
  of.ui.dotframe = CreateFrame("Frame","DotFrame",UIParent) ;
  of.ui.dotframe:SetScript( "OnUpdate", function(self, elapsed) of.on_timer( GetTime() ) ; end ) ;
  local f = of.ui.dotframe ;
  hooksecurefunc( f, "SetScale", function( self, scale )
                    if (self.fixingScale) then return end
                    self.fixingScale = true ;
                    -- check the size/scale and change them if desired
                    self.fixingScale = false
                  end
                ) ;
  f:SetAlpha( 1.0 ) ;

  of.ui.dotframe:SetSize( 107*OF.pixel_cx, 2 ) ;
  of.moveto( of.ui.dotframe, 0, 0 ) ;

  of.ui.dotframe.dots = {} ;
  local dx, dy, spacer ;
  local prev = nil ;
  dx = OF.pixel_cx ;
  dy = OF.pixel_cy ;
  spacer = 0 ;
  for i=0,106,1 do
    of.ui.dotframe.dots[i] = of.dots.create_dot( of.ui.dotframe,  i*(dx+spacer), 0, dx, dy, prev ) ;
    prev = of.ui.dotframe.dots[i] ;
  end
  of.ui.dotframe:Show() ;
end

function of.dots.set( ndx, r, g, b )
  local rc = 0 ;
  if (r == nil) then r = 0 ; rc = -1 ; end
  if (g == nil) then g = 0 ; rc = -1 ; end
  if (b == nil) then b = 0 ; rc = -1 ; end

  if (of.ui.dotframe.dots[ndx] ~= nil) then
  
    if (of.ui.dotframe.dots[ndx].texture.SetColorTexture == nil) then
      of.ui.dotframe.dots[ndx].texture:SetTexture( r/255.0, g/255.0, b/255.0, 1.0 ) ;
    else
      of.ui.dotframe.dots[ndx].texture:SetColorTexture( r/255.0, g/255.0, b/255.0, 1.0 ) ;
    end
  end
  return rc ;
end

function of.dots.set_n( ndx, val )
  local r ;
  local g ;
  local b ;
  b = val % 256 ;
  val = val / 256 ;
  g = val % 256 ;
  val = val / 256 ;
  r = val % 256 ;
  
  -- 1 byte per pixel
  of.dots.set( ndx+0, 0, 0, r ) ;
  of.dots.set( ndx+1, 0, 0, g ) ;
  of.dots.set( ndx+2, 0, 0, b ) ;
  
  -- 3 bytes per pixel
--  of.dots.set( ndx, r, g, b ) ; 
end

local dot_counter = 0 ;
function of.dots.xmit( msgid, str )
  if (_adjust) then
    -- static test message
    if (_adjust == 9) then
      msgid = 0xfffffe ;
      -- actual max msg is about 300 characters, but this should be good enough to determine whether or not we're dialed in
      str = "123456789a123456789b123456789c123456789d123456789e123456789f123456789g123456789h123456789i" ;
    elseif (_adjust == 1) then
      for i=0,300,3 do
        of.dots.set( floor(i / 3), i, i+1, i+2 ) ;
      end
      return ;
    elseif (_adjust == 2) then
      for i=0,300,3 do
        of.dots.set( floor(i / 3), 1, 100, 200 ) ;
      end
      return ;
    elseif (_adjust == 3) then
      for i=0,300,3 do
        of.dots.set( floor(i / 3), 200, 100, 200 ) ;
      end
      return ;
    elseif (_adjust == 4) then
      for i=0,300,3 do
        of.dots.set( floor(i / 3), 200, 100, 1 ) ;
      end
      return ;
    elseif (_adjust == 5) then
      for i=0,300,3 do
        of.dots.set( floor(i / 3), 200, 0, 0 ) ;
      end
      return ;
    elseif (_adjust == 6) then
      for i=0,300,3 do
        of.dots.set( floor(i / 3), 0, 0, 100 ) ;
      end
      return ;
    elseif (_adjust == 7) then
      for i=0,300,3 do
        of.dots.set( floor(i / 3), 100, 100, 100 ) ;
      end
      return ;
    elseif (_adjust == 8) then
      for i=0,100,1 do
        of.dots.set( i, 0, 0, 100 + (i % 11) ) ;
      end
      return ;
    end
  else
    dot_counter = dot_counter + 1
  end
--print( msgid .." ".. dot_counter .."  [".. str .."]" ) ;
  local ndx ;
  local beg = 15 ; -- 1 byte per pixel
--  local beg = 5 ; -- 3 bytes per pixel
--  local payload_pixels = 100 ; -- 3 bytes per pixel
  local rc  = 0 ;
  local n   = 1 ;
  local sig = "oFollow" ;
  local payload_pixels = 300 ; -- 1 byte per pixel
  
  --
  --  1 byte per pixel
  --
  
  of.dots.set( 0, 0, 0, string.byte( sig, 1 ) ) ;
  of.dots.set( 1, 0, 0, string.byte( sig, 2 ) ) ;
  of.dots.set( 2, 0, 0, string.byte( sig, 3 ) ) ;
  of.dots.set( 3, 0, 0, string.byte( sig, 4 ) ) ;
  of.dots.set( 4, 0, 0, string.byte( sig, 5 ) ) ;
  of.dots.set( 5, 0, 0, string.byte( sig, 6 ) ) ;
  of.dots.set( 6, 0, 0, string.byte( sig, 7 ) ) ;
  of.dots.set( 7, 0, 0, string.byte( sig, 8 ) ) ;
  of.dots.set( 8, 0, 0, string.byte( sig, 9 ) ) ;

  of.dots.set_n(  9, dot_counter ) ;
  of.dots.set_n( 12, msgid ) ; 
  -- string, 3 characters per dot
  
  n = 1 ;
  for ndx=0,payload_pixels,1 do
    if (rc == 0) then
      rc = of.dots.set( ndx+beg, 0, 0, string.byte( str, n )) ;
    else
      of.dots.set( ndx+beg, 0, 0, 0 ) ;
    end
    n = n + 1 ;
  end

--[[  
  --
  --  3 bytes per pixel
  --
  of.dots.set( 0, string.byte( sig, 1 ), string.byte( sig, 2 ),string.byte( sig, 3 ) ) ;
  of.dots.set( 1, string.byte( sig, 4 ), string.byte( sig, 5 ),string.byte( sig, 6 ) ) ;
  of.dots.set( 2, string.byte( sig, 7 ), string.byte( sig, 8 ),string.byte( sig, 9 ) ) ;

  of.dots.set_n( 3, dot_counter ) ;
  of.dots.set_n( 4, msgid ) ; 
  -- string, 3 characters per dot
  n = 1 ;
  for ndx=0,payload_pixels,1 do
    if (rc == 0) then
      rc = of.dots.set( ndx+5, string.byte( str, n ), string.byte( str, n+1 ),string.byte( str, n+2 )) ;
    else
      of.dots.set( ndx+5, 0, 0, 0 ) ;
    end
    n = n + 3 ;
  end
]]--

end

--------------------------------------------------------------------------
-- dots (end)
--------------------------------------------------------------------------

--------------------------------------------------------------------------
-- timer functions
--------------------------------------------------------------------------
function of.timer( id, dt_, func_, repeater, arg1_, arg2_, arg3_, arg4_, arg5_, arg6_, arg7_ )
  if (func_ == nil) then
    of.timers[ id ] = nil ;
  else
    local t = GetTime() + dt_ ;
    of.timers[ id ] = { dt = dt_, tm = t, one_shot = (not repeater), func = func_, 
                        arg1 = arg1_, arg2 = arg2_, arg3 = arg3_, arg4 = arg4_,
                        arg5 = arg5_, arg6 = arg6_, arg7 = arg7_ } ;
  end
end

function of.timer_clear()
  of.timers = nil ;
  of.timers = {} ;
end

function of.timer_dump() 
  print( "--[ timers ]------" ) ;
  for i,v in pairs( of.timers ) do
    print( "  ".. tostring(i) ..".  dt: ".. (v.dt) .."   one_shot: ".. tostring(v.one_shot) ) ;
  end  
  print( "--" ) ;
end

local one_shot = 0 ;
function of.timer_oneshot( dt_, func_, arg1_, arg2_, arg3_, arg4_, arg5_, arg6_, arg7_ )
  one_shot = one_shot + 1 ;
  of.timer( "one_shot.".. one_shot, dt_, func_, nil, arg1_, arg2_, arg3_, arg4_, arg5_, arg6_, arg7_ ) ;  
end

-- resets timer to now + dt
function of.timer_reset( id, dt )
  if (of.timers[ id ] ~= nil) then
    local now = GetTime() ;
    if (dt == nil) then
      dt = of.timers[ id ].dt ;
    end
    of.timers[ id ].tm = now + dt ;
  end
end

function of.timer_trigger( now )
  for i,v in pairs( of.timers ) do
    if (v.tm < now) then
      local arg1 = v.arg1 ;
      if (arg1 == nil) or (arg1 == "#now") then
        arg1 = now ;
      end
      local retOK, rc = pcall( v.func, arg1, v.arg2, v.arg3, v.arg4, v.arg5, v.arg6, v.arg7 ) ;
      if (retOK == true) then 
        if (rc ~= nil) or (v.one_shot) then
          of.timers[i] = nil ;
        else
          v.error_cnt = nil ;
          v.tm = now + v.dt ;
        end
      else
        v.error_cnt = (v.error_cnt or 0) + 1 ;
        if (v.error_cnt > 5) then
          print( "[of.timer] error calling '".. tostring(i) .."'.  rc[".. tostring(rc) .."].  removing timer" ) ;
          of.timers[i] = nil ;
        end
      end
    end
  end
end

function of.timer_trip( id )
  if (of.timers[ id ] ~= nil) then
    of.timers[ id ].tm = 0 ;
  end
end

--------------------------------------------------------------------------
-- timer functions (end)
--------------------------------------------------------------------------

